/***********************************************************************
*
* mimedefang-multiplexor.c
*
* Master program which manages a pool of e-mail scanning processes.
*
* Copyright (C) 2001 Roaring Penguin Software Inc.
* http://www.roaringpenguin.com
*
* This program may be distributed under the terms of the GNU General
* Public License, Version 2, or (at your option) any later version.
*
***********************************************************************/

static char const RCSID[] =
"$Id: mimedefang-multiplexor.c,v 1.66 2002/05/29 19:01:19 dfs Exp $";

#include "config.h"
#include "event_tcp.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdarg.h>
#include <pwd.h>

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif

#ifndef SUN_LEN
#define SUN_LEN(ptr)  ((size_t) (((struct sockaddr_un *) 0)->sun_path) \
       + strlen ((ptr)->sun_path))
#endif

#define MAX_CMD_LEN 512	/* Maximum length of command from mimedefang */

#define DOLOG Settings.doSyslog

#define SLAVENO(s) ((int) ((s) - AllSlaves))

/* Structure of a slave process */
typedef struct Slave_t {
    struct Slave_t *next;	/* Link in free/busy list                   */
    EventSelector *es;		/* Event selector                           */
    pid_t pid;			/* Process ID of slave process              */
    int numRequests;		/* Number of requests handled by slave      */
    time_t idleTime;		/* Time when slave was last made idle       */
    int slaveStdin;		/* Slave's stdin descriptor                 */
    int slaveStdout;		/* Slave's stdout descriptor                */
    int slaveStderr;		/* Slave's stderr descriptor                */
    int clientFD;		/* Client file descriptor                   */
    int killed;			/* Was slave killed by multiplexor?         */
    EventTcpState *event;	/* Pending event handler                    */
    EventHandler *errHandler;	/* Read handler for stderr                  */
    char workdir[MAX_CMD_LEN];	/* Working directory for current scan       */
    int generation;		/* Slave's generation                       */
} Slave;

int NumFreeSlaves;
Slave *FreeSlaveList;		/* List of available slaves                 */
Slave *AllSlaves;		/* Array of all slaves                      */
int NumBusySlaves;
Slave *BusySlaveList;		/* List of busy slaves                      */

int Generation = 0;             /* Current generation                       */
int NumRunningSlaves;		/* Number of running slave processes        */

struct Settings_t {
    int minSlaves;		/* Minimum number of slaves to keep running */
    int maxSlaves;		/* Maximum possible number of slaves        */
    int maxRequests;		/* Maximum number of requests per slave     */
    int maxIdleTime;		/* Excess slaves should be killed after time */
    int busyTimeout;		/* Timeout after which we kill scanner      */
    int clientTimeout;		/* Timeout for client request/reply         */
    int slewTime;		/* Time to wait between slaves' activation  */
    int waitTime;               /* Minimum time to wait between activations */
    int doSyslog;		/* If true, log various things with syslog  */
    char const *sockName;	/* Socket name for talking to mimedefang    */
    char const *progPath;	/* Program to execute for filter            */
    char const *statsFile;	/* File name for logging statistics         */
    char const *subFilter;      /* Sub-filter to pass to filter             */
    FILE *statsFP;		/* File pointer for stats file              */
    int statsToSyslog;		/* If true, log stats using syslog          */
    int flushStats;             /* If non-zero, flush stats file after write*/
} Settings;

/* Pipe written on reception of SIGCHLD */
static int Pipe[2] = {-1, -1};
static int ReapPending = 0;
static int HupPending = 0;
static int IntPending = 0;
static int CharPending = 0;

static int DebugEvents = 0;
static time_t LastSlaveActivation = (time_t) 0;

extern int drop_privs(char const *user, uid_t uid, gid_t gid);

/* Prototypes */
static void shutDescriptors(Slave *s);
static void reapTerminatedSlaves(void);
static pid_t activateSlave(Slave *s, char const *reason);
static void killSlave(Slave *s, char const *reason);
static void putOnBusyList(Slave *s);
static void putOnFreeList(Slave *s);
static void handleAccept(EventSelector *es, int fd);
static void handleCommand(EventSelector *es, int fd,
			  char *buf, int len, int flag, void *data);
static void handleSlaveReceivedCommand(EventSelector *es, int fd,
				       char *buf, int len, int flag,
				       void *data);
static void handleSlaveReceivedAnswer(EventSelector *es, int fd,
				      char *buf, int len, int flag,
				      void *data);
static void doScan(EventSelector *es, int fd, char *cmd);
static void doSlaveCommand(EventSelector *es, int fd, char *cmd);
static void checkSlaveForExpiry(Slave *s);
static void handlePipe(EventSelector *es,
		       int fd, unsigned int flags, void *data);
static void handleSlaveStderr(EventSelector *es,
			      int fd,
			      unsigned int flags,
			      void *data);
static void childHandler(int sig);
static void hupHandler(int sig);
static void intHandler(int sig);
static void sigterm(int sig);
static void newGeneration(void);

static void handleIdleTimeout(EventSelector *es, int fd, unsigned int flags,
			      void *data);

static void logSlaveReaped(Slave *s, int status, int busy);

static void statsReopenFile(void);
static void statsLog(char const *event, int slaveno, char const *fmt, ...);
static void bringSlavesUpToMin(EventSelector *es, int fd, unsigned int flags,
			       void *data);
static void scheduleBringSlavesUpToMin(EventSelector *es);
static int minScheduled = 0;

/**********************************************************************
* %FUNCTION: usage
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing (exits)
* %DESCRIPTION:
*  Prints usage information
***********************************************************************/
static void
usage(void)
{
    fprintf(stderr, "mimedefang-multiplexor version %s\n", VERSION);
    fprintf(stderr, "Usage: mimdefang-multiplexor [options]\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -h                -- Print usage info and exit\n");
    fprintf(stderr, "  -t filename       -- Log statistics to filename\n");
    fprintf(stderr, "  -p filename       -- Write process-ID in filename\n");
    fprintf(stderr, "  -T                -- Log statistics to syslog\n");
    fprintf(stderr, "  -u                -- Flush stats file after each write\n");
    fprintf(stderr, "  -U username       -- Run as username, not root\n");
    fprintf(stderr, "  -m minSlaves      -- Minimum number of slaves\n");
    fprintf(stderr, "  -x maxSlaves      -- Maximum number of slaves\n");
    fprintf(stderr, "  -r maxRequests    -- Maximum number of requests per slave\n");
    fprintf(stderr, "  -i idleTime       -- Idle time (seconds) for killing excess slaves\n");
    fprintf(stderr, "  -b busyTime       -- Busy time (seconds) for killing hung slaves\n");
    fprintf(stderr, "  -c cmdTime        -- Request/reply transmission timeout (seconds)\n");
    fprintf(stderr, "  -w waitTime       -- How long to wait between slave activations (seconds)\n");
    fprintf(stderr, "  -W waitTime       -- Absolute minimum to wait between slave activations\n");
    fprintf(stderr, "  -s sock           -- UNIX-domain socket for communication\n");
    fprintf(stderr, "  -f /dir/filter    -- Specify full path of filter program\n");
    fprintf(stderr, "  -d                -- Debug events in /var/log/mdefang-event-debug.log\n");
    fprintf(stderr, "  -l                -- Log events with syslog\n");
    exit(EXIT_FAILURE);
}

/**********************************************************************
* %FUNCTION: main
* %ARGUMENTS:
*  argc, argv -- usual suspects
* %RETURNS:
* Nothing -- runs in an infinite loop
* %DESCRIPTION:
*  Main program
***********************************************************************/
int
main(int argc, char *argv[])
{
    int i;
    int sock;
    int c;
    FILE *fp;
    char *pidfile = NULL;
    char *user = NULL;

    struct sockaddr_un addr;
    EventSelector *es;
    struct sigaction act;
    struct timeval t;
    struct passwd *pw = NULL;

    /* Paranoia time */
    umask(077);

    /* Paranoia time II */
    if (getuid() != geteuid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run suid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    if (getgid() != getegid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run sgid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    Settings.minSlaves     = 0;
    Settings.maxSlaves     = 2;
    Settings.maxRequests   = 100;
    Settings.maxIdleTime   = 300;
    Settings.busyTimeout   = 120;
    Settings.slewTime      = 3;
    Settings.waitTime      = 0;
    Settings.clientTimeout = 10;
    Settings.doSyslog      = 0;
    Settings.sockName      = SPOOLDIR "/mdefang-mx.sock";
    Settings.progPath      = MIMEDEFANG_PL;
    Settings.subFilter     = NULL;
    Settings.statsFile     = NULL;
    Settings.statsFP       = NULL;
    Settings.flushStats    = 0;
    Settings.statsToSyslog = 0;

    NumBusySlaves = 0;
    NumRunningSlaves = 0;
    BusySlaveList = NULL;

    while((c = getopt(argc, argv, "Tt:um:x:r:i:b:c:s:hdlf:p:w:F:W:U:")) != -1) {
	switch(c) {
	case 'U':
	    /* User to run as */
	    if (user) {
		free(user);
	    }
	    user = strdup(optarg);
	    if (!user) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'F':
	    /* Sub-filter */
	    if (Settings.subFilter) {
		free((void *) Settings.subFilter);
	    }
	    Settings.subFilter = strdup(optarg);
	    if (!Settings.subFilter) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'W':
	    /* Absolute minimum to wait between each slave's start-up */
	    if (sscanf(optarg, "%d", &Settings.waitTime) != 1) usage();
	    if (Settings.waitTime < 0) {
		Settings.waitTime = 0;
	    }
	    break;

	case 'w':
	    /* How long to wait between each slave's start-up */
	    if (sscanf(optarg, "%d", &Settings.slewTime) != 1) usage();
	    if (Settings.slewTime < 1) {
		Settings.slewTime = 1;
	    }
	    break;

	case 'p':
	    /* Write our pid to this file */
	    if (pidfile != NULL) free(pidfile);

	    pidfile = strdup(optarg);
	    if (!pidfile) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'f':
	    /* Filter program */
	    if (optarg[0] != '/') {
		fprintf(stderr, "-f: You must supply an absolute path for filter program\n");
		exit(EXIT_FAILURE);
	    }
	    Settings.progPath = strdup(optarg);
	    if (!Settings.progPath) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'u':
	    Settings.flushStats = 1;
	    break;
	case 't':
	    Settings.statsFile = strdup(optarg);
	    if (!Settings.statsFile) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'T':
	    Settings.statsToSyslog = 1;
	    break;
	case 'l':
	    Settings.doSyslog = 1;
	    break;
	case 'd':
	    DebugEvents = 1;
	    break;
	case 'h':
	    usage();
	    break;
	case 'm':
	    if (sscanf(optarg, "%d", &Settings.minSlaves) != 1) usage();
	    if (Settings.minSlaves < 0) Settings.minSlaves = 0;
	    break;
	case 'x':
	    if (sscanf(optarg, "%d", &Settings.maxSlaves) != 1) usage();
	    break;
	case 'r':
	    if (sscanf(optarg, "%d", &Settings.maxRequests) != 1) usage();
	    if (Settings.maxRequests < 1) Settings.maxRequests = 1;
	    break;
	case 'i':
	    if (sscanf(optarg, "%d", &Settings.maxIdleTime) != 1) usage();
	    if (Settings.maxIdleTime < 10) Settings.maxIdleTime = 10;
	    break;
	case 'b':
	    if (sscanf(optarg, "%d", &Settings.busyTimeout) != 1) usage();
	    if (Settings.busyTimeout < 10) Settings.busyTimeout = 10;
	    break;
	case 'c':
	    if (sscanf(optarg, "%d", &Settings.clientTimeout) != 1) usage();
	    if (Settings.clientTimeout < 10) Settings.clientTimeout = 10;
	    break;
	case 's':
	    Settings.sockName = strdup(optarg);
	    if (!Settings.sockName) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	default:
	    fprintf(stderr, "\n");
	    usage();
	}
    }

    /* Drop privileges */
    if (user) {
	fprintf(stderr, "User is: %s\n", user);
	pw = getpwnam(user);
	if (!pw) {
	    fprintf(stderr, "Unknown user '%s'\n", user);
	    exit(EXIT_FAILURE);
	}
	if (drop_privs(user, pw->pw_uid, pw->pw_gid) < 0) {
	    exit(EXIT_FAILURE);
	}
	free(user);
    }

    /* Warn */
    if (!getuid() || !geteuid()) {
	fprintf(stderr,
		"WARNING: You are running mimedefang-multiplexor as root.\n"
		"Future versions of MIMEDefang may refuse to run as root.\n");
    }

    /* Fix obvious stupidities */
    if (Settings.maxSlaves < 1) {
	Settings.maxSlaves = 1;
    }
    if (Settings.minSlaves > Settings.maxSlaves) {
	Settings.minSlaves = Settings.maxSlaves;
    }

    /* Allocate slaves and place them on the free list */
    FreeSlaveList = calloc(Settings.maxSlaves, sizeof(Slave));
    AllSlaves = FreeSlaveList;

    if (!FreeSlaveList) {
	fprintf(stderr, "Unable to allocate memory for slaves\n");
	exit(EXIT_FAILURE);
    }
    NumFreeSlaves = Settings.maxSlaves;

    /* Make an event selector */
    es = Event_CreateSelector();
    if (!es) {
	fprintf(stderr, "Could not create event selector: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Set some fields in slaves */
    for (i=0; i<NumFreeSlaves; i++) {
	Slave *s = &FreeSlaveList[i];
	s->es = es;
	s->pid = (pid_t) -1;
	s->slaveStdin = -1;
	s->slaveStdout = -1;
	s->slaveStderr = -1;
	s->clientFD = -1;
	s->killed = 0;
	s->event = NULL;
	s->errHandler = NULL;
	s->workdir[0] = 0;
	s->generation = Generation;
    }

    /* Set up the linked list */
    for (i=0; i<NumFreeSlaves-1; i++) {
	FreeSlaveList[i].next = &FreeSlaveList[i+1];
    }
    FreeSlaveList[NumFreeSlaves-1].next = NULL;

    /* Make a UNIX-domain socket to talk to mimedefang */
    (void) remove(Settings.sockName);

    sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    if (sock < 0) {
	fprintf(stderr, "Could not create UNIX-domain socket: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, Settings.sockName, sizeof(addr.sun_path)-1);

    if (bind(sock, (struct sockaddr *) &addr, SUN_LEN(&addr)) < 0) {
	fprintf(stderr, "Could not bind UNIX-domain socket %s: %s\n",
		Settings.sockName, strerror(errno));
	exit(EXIT_FAILURE);
    }
    (void) chmod(Settings.sockName, 0600);
    if (listen(sock, 5) < 0) {
	fprintf(stderr, "Could not listen on UNIX-domain socket %s: %s\n",
		Settings.sockName, strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Set up an accept loop */
    if (!EventTcp_CreateAcceptor(es, sock, handleAccept)) {
	fprintf(stderr, "Could not make accept() handler: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Ignore sigpipe */
    signal(SIGPIPE, SIG_IGN);

    /* Set up pipe for signal handler for slave termination notification*/
    if (pipe(Pipe) < 0) {
	perror("pipe");
	exit(EXIT_FAILURE);
    }

    /* Create event handler for pipe */
    if (!Event_AddHandler(es, Pipe[0],
			  EVENT_FLAG_READABLE, handlePipe, NULL)) {
	fprintf(stderr, "Could not make pipe handler: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Daemonize */
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	/* parent */
	exit(EXIT_SUCCESS);
    }
    setsid();
    signal(SIGHUP, SIG_IGN);
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	exit(EXIT_SUCCESS);
    }

    /* Write pid */
    if (pidfile) {
	fp = fopen(pidfile, "w");
	if (fp) {
	    fprintf(fp, "%d\n", (int) getpid());
	    fclose(fp);
	}
	free(pidfile);
    }

    chdir("/");
    for (i=0; i<256; i++) {
	if (i == sock || i == Pipe[0] || i == Pipe[1]) continue;
	(void) close(i);
    }

    /* Direct stdin/stdout/stderr to /dev/null */
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);

    /* Syslog if required */
    openlog("mimedefang-multiplexor", 0, LOG_MAIL);

    /* Warn in syslog also*/
    if (!getuid() || !geteuid()) {
	syslog(LOG_ERR, "WARNING: You are running mimedefang-multiplexor as root.  Future versions of MIMEDefang may refuse to run as root.");
    }

    /* Set signal handler for SIGCHLD */
    act.sa_handler = childHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    if (sigaction(SIGCHLD, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    /* Set up SIGHUP handler */
    act.sa_handler = hupHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    if (sigaction(SIGHUP, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    /* Set up SIGINT handler */
    act.sa_handler = intHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    if (sigaction(SIGINT, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    if (DOLOG) {
	syslog(LOG_INFO, "started; minSlaves=%d, maxSlaves=%d, maxRequests=%d, maxIdleTime=%d, busyTimeout=%d, clientTimeout=%d",
	       Settings.minSlaves,
	       Settings.maxSlaves,
	       Settings.maxRequests,
	       Settings.maxIdleTime,
	       Settings.busyTimeout,
	       Settings.clientTimeout);
    }

    /* Open stats file */
    statsReopenFile();

    /* Kick off the starting of slaves */
    bringSlavesUpToMin(es, 0, 0, NULL);

    /* Set up a timer handler to check for idle timeouts */
    t.tv_usec = 0;
    t.tv_sec = Settings.maxIdleTime;
    Event_AddTimerHandler(es, t, handleIdleTimeout, NULL);

    if (DebugEvents) {
	Event_EnableDebugging("/var/log/mdefang-event-debug.log");
    }

    /* Set signal handler for SIGTERM */
    signal(SIGTERM, sigterm);

    /* And loop... */
    while(1) {
	if (Event_HandleEvent(es) < 0) {
	    syslog(LOG_CRIT, "Error in Event_HandleEvent: %m.  MULTIPLEXOR IS TERMINATING.");
	    sigterm(0);
	}
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveStderr
* %ARGUMENTS:
*  es -- event selector
*  fd -- file descriptor ready for reading
*  flags -- flags from event-handling code
*  data -- pointer to slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a readable stderr descriptor.  Reads data; if we're logging,
*  sends it to syslog; otherwise, discards data.
***********************************************************************/
static void
handleSlaveStderr(EventSelector *es,
		  int fd,
		  unsigned int flags,
		  void *data)
{
    Slave *s = (Slave *) data;
    char buffer[512];
    int n;

    while ( (n=read(fd, buffer, sizeof(buffer)-1)) > 0) {
	buffer[n] = 0;
	if (DOLOG) {
	    syslog(LOG_INFO, "Slave %d stderr: %s", SLAVENO(s), buffer);
	}
    }
    if (n == 0 || (n < 0 && errno != EAGAIN)) {
	/* EOF or error reading stderr -- close it and cancel handler */
	if (n < 0) {
	    if (DOLOG) {
		syslog(LOG_WARNING,
		       "handleSlaveStderr: Error reading from slave %d's stderr: %m", SLAVENO(s));
	    }
	}
	close(s->slaveStderr);
	Event_DelHandler(s->es, s->errHandler);
	s->errHandler = NULL;
	s->slaveStderr = -1;
    }
}

/**********************************************************************
* %FUNCTION: handleAccept
* %ARGUMENTS:
*  es -- event selector
*  fd -- accepted connection
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a connection attempt from MIMEDefang.  Sets up a command-reader.
***********************************************************************/
static void
handleAccept(EventSelector *es, int fd)
{
    if (!EventTcp_ReadBuf(es, fd, MAX_CMD_LEN, '\n', handleCommand,
			  Settings.clientTimeout, NULL)) {
	if (DOLOG) {
	    syslog(LOG_ERR, "handleAccept: EventTcp_ReadBuf failed: %m");
	}
	close(fd);
    }
}

/**********************************************************************
* %FUNCTION: handleCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a MIMEDefang command.  Currently, only two commands are
*  available:
*  "free\n" --> Returns number of free slaves.
*  "scan dir\n" --> Scans directory, returns "done" when done or "busy" if
*      no free slaves or "error: msg" if an error occurred.
***********************************************************************/
static void
handleCommand(EventSelector *es,
	      int fd,
	      char *buf,
	      int len,
	      int flag,
	      void *data)
{
    char answer[MAX_CMD_LEN];

    if (flag == EVENT_TCP_FLAG_TIMEOUT || flag == EVENT_TCP_FLAG_IOERROR) {
	if (DOLOG) {
	    syslog(LOG_DEBUG, "handleCommand: Timeout or error: Flag = %d",
		   flag);
	}
	/* Client timeout or error */
	close(fd);
	return;
    }

    /* Null-terminate buffer */
    if (len) {
	buf[len-1] = 0;
	len--;
    }

    if (len == 4 && !strcmp(buf, "free")) {
	snprintf(answer, sizeof(answer), "%d\n", NumFreeSlaves);
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) {
		syslog(LOG_ERR, "handleCommand: EventTcp_WriteBuf failed: %m");
	    }
	    close(fd);
	}
	return;
    }

    if (len > 5 && !strncmp(buf, "scan ", 5)) {
	doScan(es, fd, buf);
	return;
    }

    if (len > 8 && !strncmp(buf, "relayok ", 8)) {
	doSlaveCommand(es, fd, buf);
	return;
    }
    if (len > 10 && !strncmp(buf, "senderok ", 8)) {
	doSlaveCommand(es, fd, buf);
	return;
    }

    /* Unknown command */
    if (DOLOG) {
	syslog(LOG_ERR, "Unknown command: %s", buf);
    }
    snprintf(answer, sizeof(answer), "error: Unknown command %s\n",
	     buf);
    if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			   Settings.clientTimeout, NULL)) {
	if (DOLOG) {
	    syslog(LOG_ERR, "handleCommand: EventTcp_WriteBuf failed: %m");
	}
	close(fd);
    }
}

/**********************************************************************
* %FUNCTION: doScan
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  cmd -- scan command
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends the scan command to the slave and arranges for answer to be
*  sent back when scanning is complete.
***********************************************************************/
static void
doScan(EventSelector *es, int fd, char *cmd)
{
    Slave *s;
    int len;

    /* Add newline back to command */
    len = strlen(cmd);
    if (len < MAX_CMD_LEN-1) {
	cmd[len+1] = 0;
	cmd[len] = '\n';
    } else {
	char *answer = "error: Command too long\n";
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doScan: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}

	if (DOLOG) {
	    syslog(LOG_DEBUG, "doScan: Command too long");
	}
	return;
    }

    /* Find a free slave */
    s = FreeSlaveList;
    if (!s) {
	char *answer = "error: No free slaves\n";
	if (DOLOG) {
	    syslog(LOG_WARNING, "No free slaves");
	}
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doScan: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}
	return;
    }

    if (activateSlave(s, "About to perform scan") == (pid_t) -1) {
	char *answer = "error: Unable to activate slave\n";
	if (DOLOG) {
	    syslog(LOG_ERR, "Unable to activate slave %d",
		   SLAVENO(s));
	}
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doScan: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}
	return;
    }

    /* Put the slave on the busy list */
    putOnBusyList(s);

    /* Set slave's clientFD so we can reply */
    s->clientFD = fd;

    /* Set slave's working directory, sans newline */
    snprintf(s->workdir, sizeof(s->workdir), "%s", &cmd[strlen("scan ")]);
    len = strlen(s->workdir);
    if (len && s->workdir[len-1] == '\n') {
	s->workdir[len-1] = 0;
    }

    /* And tell the slave to go ahead... */
    s->event = EventTcp_WriteBuf(es, s->slaveStdin, cmd, strlen(cmd),
				 handleSlaveReceivedCommand,
				 Settings.clientTimeout, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "doScan: EventTcp_WriteBuf failed: %m");
	killSlave(s, "EventTcp_WriteBuf failed");
    } else {
	statsLog("StartFilter", SLAVENO(s), NULL);
    }
}

/**********************************************************************
* %FUNCTION: doSlaveCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  cmd -- command
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends the command to the slave and arranges for answer to be
*  sent back when command finishes.
***********************************************************************/
static void
doSlaveCommand(EventSelector *es, int fd, char *cmd)
{
    Slave *s;
    int len;
    char reason[200];

    sprintf(reason, "About to execute command '%.100s'", cmd);

    /* Add newline back to command */
    len = strlen(cmd);
    if (len < MAX_CMD_LEN-1) {
	cmd[len+1] = 0;
	cmd[len] = '\n';
    } else {
	char *answer = "error: Command too long\n";
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doSlaveCommand: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}

	if (DOLOG) {
	    syslog(LOG_DEBUG, "doSlaveCommand: Command too long");
	}
	return;
    }

    /* Find a free slave */
    s = FreeSlaveList;
    if (!s) {
	char *answer = "error: No free slaves\n";
	if (DOLOG) {
	    syslog(LOG_WARNING, "No free slaves");
	}
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doSlaveCommand: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}
	return;
    }

    if (activateSlave(s, reason) == (pid_t) -1) {
	char *answer = "error: Unable to activate slave\n";
	if (DOLOG) {
	    syslog(LOG_ERR, "Unable to activate slave %d",
		   SLAVENO(s));
	}
	if (!EventTcp_WriteBuf(es, fd, answer, strlen(answer), NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "doSlaveCommand: EventTcp_WriteBuf failed: %m");
	    close(fd);
	}
	return;
    }

    /* Put the slave on the busy list */
    putOnBusyList(s);

    /* Set slave's clientFD so we can reply */
    s->clientFD = fd;

    /* Null workdir signals not to log EndFilter event */
    s->workdir[0] = 0;

    /* And tell the slave to go ahead... */
    s->event = EventTcp_WriteBuf(es, s->slaveStdin, cmd, strlen(cmd),
				 handleSlaveReceivedCommand,
				 Settings.clientTimeout, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "doSlaveCommand: EventTcp_WriteBuf failed: %m");
	killSlave(s, "EventTcp_WriteBuf failed");
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveReceivedCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- the slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when command has been written to slave.  Sets up an event handler
*  to read the slave's reply
***********************************************************************/
static void
handleSlaveReceivedCommand(EventSelector *es,
			   int fd,
			   char *buf,
			   int len,
			   int flag,
			   void *data)
{
    Slave *s = (Slave *) data;

    /* Event was triggered */
    s->event = NULL;

    if (flag == EVENT_TCP_FLAG_TIMEOUT || flag == EVENT_TCP_FLAG_IOERROR) {
	/* Error writing to slave */
	char *answer = "error: Error talking to slave process\n";
	if (DOLOG) {
	    syslog(LOG_DEBUG, "handleSlaveReceivedCommand(%d): Timeout or error: Flag = %d",
		   SLAVENO(s),
		   flag);
	}
	if (!EventTcp_WriteBuf(es, s->clientFD, answer, strlen(answer),
			       NULL, Settings.clientTimeout, NULL)) {
	    if (DOLOG) syslog(LOG_ERR, "handleSlaveReceivedCommand: EventTcp_WriteBuf failed: %m");
	}
	/* Kill the slave process */
	killSlave(s, "Error talking to slave process");
	putOnFreeList(s);
	return;
    }

    /* Slave has been given the command; now wait for it to reply */
    s->event = EventTcp_ReadBuf(es, s->slaveStdout, MAX_CMD_LEN, '\n',
				handleSlaveReceivedAnswer,
				Settings.busyTimeout, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "handleSlaveReceivedCommand: EventTcp_WriteBuf failed: %m");
	killSlave(s, "EventTcp_ReadBuf failed");
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveReceivedAnswer
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- the slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when the slave's answer comes back.
***********************************************************************/
static void
handleSlaveReceivedAnswer(EventSelector *es,
			  int fd,
			  char *buf,
			  int len,
			  int flag,
			  void *data)
{
    Slave *s = (Slave *) data;
    EventTcpState *r;

    /* Event was triggered */
    s->event = NULL;

    /* If nothing was received from slave, send error message back */
    if (!len || (flag == EVENT_TCP_FLAG_TIMEOUT)) {
	if (flag == EVENT_TCP_FLAG_TIMEOUT) {
	    /* Heuristic... */
	    if (NumBusySlaves > 3) {
		r = EventTcp_WriteBuf(es, s->clientFD,
				      "ERR Filter timed out - system may be overloaded "
				      "(consider increasing busy timeout)\n",
				      strlen("ERR Filter timed out - system may be overloaded "
					     "(consider increasing busy timeout)\n"),
				      NULL,
				      Settings.clientTimeout, NULL);
	    } else {
		r = EventTcp_WriteBuf(es, s->clientFD, "ERR Filter timed out - check filter rules or system load\n",
				      strlen("ERR Filter timed out - check filter rules or system load\n"), NULL,
				      Settings.clientTimeout, NULL);
	    }
	} else {
	    if (DOLOG) {
		syslog(LOG_ERR, "Slave died prematurely -- check your filter rules");
	    } else {
		syslog(LOG_ERR, "Slave died prematurely -- check your filter rules and use the '-l' flag on mimedefang-multiplexor to see Perl error messages");
	    }
	    r = EventTcp_WriteBuf(es, s->clientFD, "ERR No response from slave\n",
				  strlen("ERR No response from slave\n"), NULL,
				  Settings.clientTimeout, NULL);
	}
	if (!r) {
	    if (DOLOG) {
		syslog(LOG_ERR, "handleSlaveReceivedAnswer: EventTcp_WriteBuf failed: %m");
	    }
	    close(s->clientFD);
	}
    } else {
	/* Write the slave's answer back to the client */
	if (!EventTcp_WriteBuf(es, s->clientFD, buf, len, NULL,
			       Settings.clientTimeout, NULL)) {
	    if (DOLOG) {
		syslog(LOG_ERR, "handleSlaveReceivedAnswer: EventTcp_WriteBuf failed: %m");
	    }
	    close(s->clientFD);
	}
    }

    /* Put slave on free list */
    putOnFreeList(s);


    if (s->workdir[0]) {
	/* Update number of requests */
	s->numRequests++;
	statsLog("EndFilter", SLAVENO(s), "numRequests=%d", s->numRequests);
    }

    /* If we had a busy timeout, kill the slave */
    if (flag == EVENT_TCP_FLAG_TIMEOUT) {
	killSlave(s, "Busy timeout");
    } else {
	/* Check slave for expiry */
	checkSlaveForExpiry(s);
    }
}

/**********************************************************************
* %FUNCTION: putOnFreeList
* %ARGUMENTS:
*  s -- a slave which is currently on the busy list
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Puts a slave on the free list.  The slave MUST currently be on the
*  busy list.
* %PRECONDITIONS:
*  s must be on the busy list.
***********************************************************************/
static void
putOnFreeList(Slave *s)
{
    Slave *prev = NULL;
    if (s != BusySlaveList) {
	for (prev = BusySlaveList; prev && prev->next != s; prev = prev->next);
	if (!prev) {
	    if (DOLOG) {
		syslog(LOG_ERR, "putOnFreeList(%d): Slave not on busy list",
		       SLAVENO(s));
	    }
	    /* Error!  Not on busy list */
	    return;
	}
    }

    /* Remove from busy list */
    if (prev) {
	prev->next = s->next;
    } else {
	BusySlaveList = s->next;
    }

    /* Add to free list */
    s->next = FreeSlaveList;
    FreeSlaveList = s;

    /* Adjust counters */
    NumBusySlaves--;
    NumFreeSlaves++;

    /* Record time when this slave became idle */
    s->idleTime = time(NULL);

    /* Disconnect from client */
    s->event = NULL;
    s->clientFD = -1;
}

/**********************************************************************
* %FUNCTION: putOnBusyList
* %ARGUMENTS:
*  s -- a slave which is currently on the free list
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Puts a slave on the free list.  The slave MUST currently be on the
*  free list.
* %PRECONDITIONS:
*  s must be on the free list.
***********************************************************************/
static void
putOnBusyList(Slave *s)
{
    Slave *prev = NULL;
    if (s != FreeSlaveList) {
	for (prev = FreeSlaveList; prev && prev->next != s; prev = prev->next);
	if (!prev) {
	    if (DOLOG) {
		syslog(LOG_ERR, "putOnBusyList(%d): Slave not on free list",
		       SLAVENO(s));
	    }
	    return;
	}
    }

    /* Remove from free list */
    if (prev) {
	prev->next = s->next;
    } else {
	FreeSlaveList = s->next;
    }

    /* Add to busy list */
    s->next = BusySlaveList;
    BusySlaveList = s;

    /* Adjust counters */
    NumFreeSlaves--;
    NumBusySlaves++;

    /*
        if (DOLOG) {
	    syslog(LOG_DEBUG, "putOnBusyList(%d): Busy=%d, free=%d",
	           SLAVENO(s),
	           NumBusySlaves, NumFreeSlaves);
        }
    */

}

/**********************************************************************
* %FUNCTION: activateSlave
* %ARGUMENTS:
*  s -- a slave
*  reason -- reason slave is being activated
* %RETURNS:
*  The process-ID of the slave
* %DESCRIPTION:
*  Activates the slave if it is not currently associated with a running
*  process.
***********************************************************************/
static pid_t
activateSlave(Slave *s, char const *reason)
{
    int pin[2], pout[2], perr[2];
    char const *pname;
    int i;
    int flags;
    time_t now = (time_t) 0; /* Avoid compiler warning by initializing */

    /* Check if it's already active */
    if (s->pid != (pid_t) -1) return s->pid;

    /* Check if enough time has elapsed */
    if (Settings.waitTime) {
	now = time(NULL);
	if (LastSlaveActivation &&
	     (((int) (now - LastSlaveActivation)) < Settings.waitTime)) {
	    if (DOLOG) {
		syslog(LOG_WARNING, "Did not start slave %d: Not enough time elapsed since last slave activation", SLAVENO(s));
	    }
	    return -1;
	}
    }

    /* Set up pipes */
    if (pipe(pin) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	return -1;
    }
    if (pipe(pout) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	return -1;
    }

    if (pipe(perr) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	close(pout[0]);
	close(pout[1]);
	return -1;
    }

    /* fork and exec */
    s->pid = fork();

    if (s->pid == (pid_t) -1) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: fork failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	close(pout[0]);
	close(pout[1]);
	close(perr[0]);
	close(perr[1]);
	return s->pid; /* Fork failed */
    }

    if (s->pid) {
	NumRunningSlaves++;

	/* In the parent -- return */
	close(pin[0]);
	close(pout[1]);
	close(perr[1]);
	s->slaveStdin = pin[1];
	s->slaveStdout = pout[0];
	s->slaveStderr = perr[0];

	/* Make slaveStderr non-blocking */
	flags = fcntl(s->slaveStderr, F_GETFL, 0);
	if (flags != -1) {
	    flags = fcntl(s->slaveStderr, F_SETFL, flags | O_NONBLOCK);
	}

	if (flags == -1) {
	    /* Doh.. next to impossible to recover from this error */
	    syslog(LOG_ERR, "Could not make slave's stderr non-blocking: %m");
	}

	/* Handle anything written to slave's stderr */
	s->errHandler = Event_AddHandler(s->es, s->slaveStderr,
					 EVENT_FLAG_READABLE,
					 handleSlaveStderr, s);
	s->clientFD = -1;
	s->numRequests = 0;
	s->killed = 0;
	s->generation = Generation;
	if (DOLOG) {
	    syslog(LOG_INFO, "Starting slave %d (pid %lu) (%d running): %s",
		   SLAVENO(s),
		   (unsigned long) s->pid,
		   NumRunningSlaves, reason);
	}
	statsLog("StartSlave", SLAVENO(s), "reason=\"%s\"", reason);
	if (Settings.waitTime) {
	    LastSlaveActivation = now;
	}
	return s->pid;
    }

    /* In the child -- exec the program */
    close(pin[1]);
    close(pout[0]);
    close(perr[0]);
    dup2(pin[0], STDIN_FILENO);
    dup2(pout[1], STDOUT_FILENO);
    dup2(perr[1], STDERR_FILENO);

    if (pin[0] != STDIN_FILENO) close(pin[0]);
    if (pout[1] != STDOUT_FILENO) close(pout[1]);
    if (perr[1] != STDERR_FILENO) close(perr[1]);

    /* Close unneeded file descriptors */
    for (i=STDERR_FILENO+1; i<256; i++) {
	(void) close(i);
    }

    pname = strrchr(Settings.progPath, '/');
    if (pname) {
	pname++;
    } else {
	pname = Settings.progPath;
    }
    if (Settings.subFilter) {
	execl(Settings.progPath, pname, "-f", Settings.subFilter, "-server", NULL);
    } else {
	execl(Settings.progPath, pname, "-server", NULL);
    }
    _exit(EXIT_FAILURE);
}

/**********************************************************************
* %FUNCTION: checkSlaveForExpiry
* %ARGUMENTS:
*  s -- a slave to check
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  If the slave has served too many requests, it is killed.
***********************************************************************/
static void
checkSlaveForExpiry(Slave *s)
{
    if (s->numRequests > Settings.maxRequests) {
	killSlave(s, "Slave has processed maxRequests requests");
    } else if (s->generation < Generation) {
	killSlave(s, "New generation -- forcing reread of filter rules");
    }
}

/**********************************************************************
* %FUNCTION: killSlave
* %ARGUMENTS:
*  s -- a slave to kill
*  reason -- reason slave is being killed
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Kills a slave
***********************************************************************/
static void
killSlave(Slave *s, char const *reason)
{
    if (s->pid != (pid_t) -1) {
	if (DOLOG) syslog(LOG_INFO, "Killing slave %d (pid %lu): %s",
			  SLAVENO(s),
			  (unsigned long) s->pid, reason);
	/* In case, for some weird reason, the slave has stopped... */
	kill(s->pid, SIGCONT);
	kill(s->pid, SIGTERM);
	s->killed = 1;
	statsLog("KillSlave", SLAVENO(s), "reason=\"%s\"", reason);
    }
}

/**********************************************************************
*%FUNCTION: childHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGCHLD.  Writes 'C' to Pipe to wake up the select
* loop and cause reaping of dead sessions
***********************************************************************/
static void
childHandler(int sig)
{
    char byte = 'C';
    if (!ReapPending) {
	ReapPending = 1;
	if (!CharPending) {
	    write(Pipe[1], &byte, 1);
	    CharPending = 1;
	}
    }
}

/**********************************************************************
*%FUNCTION: hupHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGHUP.  Writes 'H' to Pipe to wake up the select
* loop and cause closing and reopening of stats descriptor
***********************************************************************/
static void
hupHandler(int sig)
{
    char byte = 'H';
    if (!HupPending) {
	HupPending = 1;
	if (!CharPending) {
	    write(Pipe[1], &byte, 1);
	    CharPending = 1;
	}
    }
}

/**********************************************************************
*%FUNCTION: intHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGINT.  Writes 'I' to Pipe to wake up the select
* loop and cause closing and reopening of stats descriptor
***********************************************************************/
static void
intHandler(int sig)
{
    char byte = 'I';
    if (!IntPending) {
	IntPending = 1;
	if (!CharPending) {
	    write(Pipe[1], &byte, 1);
	    CharPending = 1;
	}
    }
}

/**********************************************************************
* %FUNCTION: handlePipe
* %ARGUMENTS:
*  es -- event selector (ignored)
*  fd -- file descriptor which is readable
*  flags -- ignored
*  data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Simply reads from the pipe; then reaps and/or closes and reopens
*  stats file.
***********************************************************************/
static void
handlePipe(EventSelector *es,
	    int fd,
	    unsigned int flags,
	    void *data)
{
    char buf[64];
    sigset_t sigs, osigs;
    int n;
    int doreap, dohup, doint;

    /* The read and the reset of CharPending must be atomic with respect
       to signal handlers. */
    sigemptyset(&sigs);
    sigaddset(&sigs, SIGCHLD);
    sigaddset(&sigs, SIGINT);
    sigaddset(&sigs, SIGHUP);

    n = sigprocmask(SIG_BLOCK, &sigs, &osigs);
    if (n < 0) {
	syslog(LOG_ERR, "sigprocmask failed: %m");
    }

    /* Begin atomic portion */
    read(fd, buf, 64);
    doreap = ReapPending;
    doint  = IntPending;
    dohup  = HupPending;
    CharPending = 0;
    ReapPending = 0;
    IntPending  = 0;
    HupPending  = 0;
    /* End atomic portion */

    /* Reset signal mask */
    if (n >= 0) {
	sigprocmask(SIG_SETMASK, &osigs, NULL);
    }

    if (doreap) {
	reapTerminatedSlaves();

	/* Activate new slaves if we've fallen below minimum */
	scheduleBringSlavesUpToMin(es);
    }
    if (dohup) {
	statsReopenFile();
    }
    if (doint) {
	newGeneration();
    }
}

/**********************************************************************
* %FUNCTION: reapTerminatedSlaves
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Reaps all terminated slaves
***********************************************************************/
static void
reapTerminatedSlaves(void)
{
    pid_t pid;
    int status;
    Slave *s;

    while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
	for (s = FreeSlaveList; s; s = s->next) {
	    if (s->pid == pid) {
		if (DOLOG) logSlaveReaped(s, status, 0);
		s->killed = 0;
		NumRunningSlaves--;
		s->pid = (pid_t) -1;
		shutDescriptors(s);
		statsLog("ReapSlave", SLAVENO(s), NULL);
		break;
	    }
	}
	for (s = BusySlaveList; s; s = s->next) {
	    if (s->pid == pid) {
		if (DOLOG) logSlaveReaped(s, status, 1);
		s->killed = 0;
		NumRunningSlaves--;
		s->pid = (pid_t) -1;
		shutDescriptors(s);
		putOnFreeList(s);
		statsLog("ReapSlave", SLAVENO(s), NULL);
		break;
	    }
	}
    }
}

/**********************************************************************
* %FUNCTION: shutDescriptors
* %ARGUMENTS:
*  es -- event selector
*  s -- a slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Closes slave's descriptors if necessary
***********************************************************************/
static void
shutDescriptors(Slave *s)
{
    if (s->slaveStdin >= 0) {
	close(s->slaveStdin);
	s->slaveStdin = -1;
    }
    if (s->slaveStdout >= 0) {
	close(s->slaveStdout);
	s->slaveStdout = -1;
    }
    if (s->slaveStderr >= 0) {
	close(s->slaveStderr);
	Event_DelHandler(s->es, s->errHandler);
	s->errHandler = NULL;
	s->slaveStderr = -1;
    }
    if (s->clientFD >= 0) {
	close(s->clientFD);
	s->clientFD = -1;
    }
    if (s->event) {
	EventTcp_CancelPending(s->event);
	s->event = NULL;
    }
}

/**********************************************************************
* %FUNCTION: handleIdleTimeout
* %ARGUMENTS:
*  es -- event selector
*  fd, flags, data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called periodically to check for idle slaves that can be killed
***********************************************************************/
static void
handleIdleTimeout(EventSelector *es,
		  int fd,
		  unsigned int flags,
		  void *data)
{
    time_t now = time(NULL);
    struct timeval t;

    /* Kill idle slaves which have been idle for maxIdleTime */
    Slave *s = FreeSlaveList;
    int numAlive = NumRunningSlaves;
    while(s && numAlive > Settings.minSlaves) {
	if (((unsigned long) now - (unsigned long) s->idleTime)
	    >= Settings.maxIdleTime) {
	    numAlive--;
	    killSlave(s, "Idle timeout");
	}
	s = s->next;
    }

    /* If there are fewer running slaves than Settings.minSlaves,
       then start some more.  Should never happen, but... */
    if (numAlive < Settings.minSlaves) {
	scheduleBringSlavesUpToMin(es);
    }

    /* Reschedule timer */
    t.tv_usec = 0;
    t.tv_sec = Settings.maxIdleTime;
    Event_AddTimerHandler(es, t, handleIdleTimeout, NULL);
}

/**********************************************************************
* %FUNCTION: logSlaveReaped
* %ARGUMENTS:
*  s -- slave
*  status -- termination status
*  busy -- 1 if slave was on busy list, 0 otherwise.
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Logs to syslog
***********************************************************************/
static void
logSlaveReaped(Slave *s, int status, int busy)
{
    int level;
    char *extra;
    char *idle;

    if (busy) {
	idle = "Busy";
    } else {
	idle = "Idle";
    }

    if (s->killed) {
	level = LOG_DEBUG;
	extra = "";
    } else {
	level = LOG_ERR;
	extra = " (SLAVE DIED UNEXPECTEDLY)";
    }
    if (WIFEXITED(status)) {
	syslog(level, "Reap: %s slave %d (pid %lu) exited normally with status %d%s",
	       idle,
	       SLAVENO(s), (unsigned long) s->pid, WEXITSTATUS(status),
	       extra);
	return;
    }
    if (WIFSIGNALED(status)) {
	if (s->killed && WTERMSIG(status) == SIGTERM) {
	    syslog(level, "Reap: %s slave %d (pid %lu) exited due to SIGTERM as expected.",
		   idle,
		   SLAVENO(s), (unsigned long) s->pid);
	} else {
	    syslog(level, "Reap: %s slave %d (pid %lu) exited due to signal %d%s",
		   idle,
		   SLAVENO(s), (unsigned long) s->pid, WTERMSIG(status),
		   extra);
	}
	return;
    }
    syslog(level, "Reap: %s slave %d (pid %lu) exited for unknown reason%s",
	   idle,
	   SLAVENO(s), (unsigned long) s->pid, extra);
}

/**********************************************************************
* %FUNCTION: sigterm
* %ARGUMENTS:
*  sig -- signal number
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when SIGTERM received -- kills slaves and exits
***********************************************************************/
static void
sigterm(int sig)
{
    int i;

    if (DOLOG) {
	if (sig) {
	    syslog(LOG_INFO, "Received SIGTERM: Killing slaves and terminating");
	}
    }
    for (i=0; i<Settings.maxSlaves; i++) {
	if (AllSlaves[i].pid != (pid_t) -1) {
	    kill(AllSlaves[i].pid, SIGCONT);
	    kill(AllSlaves[i].pid, SIGTERM);
	}
    }
    exit(1);
}

/**********************************************************************
* %FUNCTION: statsReopenFile
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Opens or re-opens statistics file.
***********************************************************************/
static void
statsReopenFile(void)
{
    if (!Settings.statsFile) return;
    if (Settings.statsFP) {
	if (fclose(Settings.statsFP) == EOF) {
	    syslog(LOG_ERR, "Failed to close stats file: %m");
	}
	Settings.statsFP = NULL;
    }

    Settings.statsFP = fopen(Settings.statsFile, "a");
    if (!Settings.statsFP) {
	syslog(LOG_ERR, "Could not open stats file %s: %m",
	       Settings.statsFile);
    }
}

/**********************************************************************
* %FUNCTION: statsLog
* %ARGUMENTS:
*  event -- name of event to put in log
*  slaveno -- slave involved in this event, or -1 if not slave-specific
*  fmt   -- "printf" format string with extra args to add to line
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Logs an event to the stats file
***********************************************************************/
static void
statsLog(char const *event, int slaveno, char const *fmt, ...)
{
    struct timeval now;
    time_t tnow;
    struct tm *t;
    char tbuf[64];
    char statbuf[1024];
    va_list ap;
    int ms;

    if (!Settings.statsFP && !Settings.statsToSyslog) return;

    gettimeofday(&now, NULL);
    tnow = (time_t) now.tv_sec;
    t = localtime(&tnow);
    ms = now.tv_usec / 1000;

    strftime(tbuf, sizeof(tbuf), "%d/%m/%Y:%H:%M:%S", t);
    snprintf(statbuf, sizeof(statbuf),
	     "%s %lu.%03d %s slave=%d nslaves=%d nbusy=%d",
	     tbuf, (unsigned long) tnow, ms, event, slaveno,
	     NumRunningSlaves, NumBusySlaves);

    statbuf[sizeof(statbuf)-1] = 0;

    if (fmt) {
	int len = strlen(statbuf);
	statbuf[len] = ' ';
	va_start(ap, fmt);
	vsnprintf(statbuf+len+1, sizeof(statbuf)-len-1, fmt, ap);
	va_end(ap);
	statbuf[sizeof(statbuf)-1] = 0;
    }

    if (Settings.statsFP) {
	fprintf(Settings.statsFP, "%s\n", statbuf);
	if (Settings.flushStats) fflush(Settings.statsFP);
    }
    if (Settings.statsToSyslog) {
	syslog(LOG_INFO, "stats %s", statbuf);
    }
}

/**********************************************************************
* %FUNCTION: newGeneration
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Kills all running-but-idle slaves and increments Generation.  This
*  causes running-but-busy slaves to terminate at the earliest opportunity.
*  Use this if you change the filter file and want to restart the slaves.
***********************************************************************/
static void
newGeneration(void)
{
    Slave *s = FreeSlaveList;

    Generation++;

    while(s) {
	killSlave(s,
		  "Received SIGINT -- forcing reread of filter rules");
	s = s->next;
    }
}

/**********************************************************************
* %FUNCTION: bringSlavesUpToMin
* %ARGUMENTS:
*  es -- event selector
*  fd, flags, data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  If there are fewer than minSlaves running, start one.  If still fewer,
*  schedule self to re-run in slewTime seconds.
***********************************************************************/
void
bringSlavesUpToMin(EventSelector *es,
		   int fd,
		   unsigned int flags,
		   void *data)
{
    Slave *s;

    minScheduled = 0;

    if (NumRunningSlaves >= Settings.minSlaves) {
	/* Enough slaves, so do nothing */
	return;
    }

    /* Start a slave */
    for (s = FreeSlaveList; s; s=s->next) {
	if (s->pid == (pid_t) -1) {
	    activateSlave(s, "Bringing slaves up to minSlaves");
	    break;
	}
    }

    /* Reschedule if necessary */
    if (NumRunningSlaves < Settings.minSlaves) {
	scheduleBringSlavesUpToMin(es);
    }
}

/**********************************************************************
* %FUNCTION: scheduleBringSlavesUpToMin
* %ARGUMENTS:
*  es -- event selector
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Schedules a call to bringSlavesUpToMin in slewTime seconds from now.
***********************************************************************/
static void
scheduleBringSlavesUpToMin(EventSelector *es)
{
    struct timeval t;

    /* Do nothing if already scheduled */
    if (minScheduled) return;

    minScheduled = 1;
    t.tv_usec = 0;
    t.tv_sec = Settings.slewTime;
    Event_AddTimerHandler(es, t, bringSlavesUpToMin, NULL);
}
