/***********************************************************************
*
* mimedefang.c
*
* C interface to the attachment-filter program for stripping or altering
* MIME attachments in incoming Sendmail connections.
*
* Copyright (C) 2000 Roaring Penguin Software Inc.
* http://www.roaringpenguin.com
*
* This program may be distributed under the terms of the GNU General
* Public License, Version 2, or (at your option) any later version.
*
* This program was derived from the sample mail filter included in
* libmilter/README in the Sendmail 8.11 distribution.
***********************************************************************/

static char const RCSID[] =
"$Id: mimedefang.c,v 1.93 2002/06/03 15:01:37 dfs Exp $";

/* Define this to work around an M$ Outlook bug! */
/* #define CONVERT_EMBEDDED_CRS_IN_HEADERS 1 */

#define _BSD_SOURCE 1

#ifdef __linux__
/* On Linux, we need this defined to get fdopen.  On BSD, if we define
 * it, we don't get all the u_char, u_long, etc definitions.  GRR! */
#define _POSIX_SOURCE 1
#endif

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#include <pwd.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>

#include "libmilter/mfapi.h"

#include <stdio.h>
#include <sys/socket.h>
#include <sys/un.h>

#ifdef ENABLE_DEBUGGING
#include <signal.h>
#define DEBUG(x) x
#else
#define DEBUG(x) (void) 0
#endif

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif


#define DEBUG_ENTER(func, line) DEBUG(syslog(LOG_INFO, "Entering %s (line %d)", func, line))
#define DEBUG_EXIT(func, line, ret) DEBUG(syslog(LOG_INFO, "Exiting %s (line %d) ret=%s", func, line, ret))

#define SCAN_BODY \
"MIMEDefang " VERSION " (www dot roaringpenguin dot com slash mimedefang)"

/* In debug mode, we do not delete working directories. */
int DebugMode = 0;

/* Run as this user */
static char *user = NULL;
extern int drop_privs(char const *user, uid_t uid, gid_t gid);

/* Keep track of private data -- file name and fp for writing e-mail body */
struct privdata {
    char *hostname;		/* Name of connecting host */
    char *hostip;		/* IP address of connecting host */
    char *dir;			/* Work directory */
    FILE *fp;			/* File for message body */
    FILE *headerFP;		/* File for message headers */
    FILE *recipFP;		/* File for recipients */
    unsigned char headerPresent; /* Already an X-Scanned-By: header found */
    unsigned char suspiciousBody; /* Suspicious characters in message body? */
    unsigned char lastWasCR;	/* Last char of body chunk was CR? */
    unsigned char filterFailed; /* Filter failed */
};

static int MXRelayOK(SMFICTX *ctx, char const *ip, char const *name);
static int MXSenderOK(SMFICTX *ctx, char const *sender);

static void chomp(char *str);
static void convert_newline_sequences(char *str);

#ifdef FOR_FUTURE_RELEASE
static void write_percent_encoded(unsigned char const *str, FILE *fp);
static void percent_decode(unsigned char *buf);
#endif

static void log_perl_stderr(SMFICTX *ctx);
static sfsistat cleanup(SMFICTX *ctx);
static int IncrementRunningFilters(void);
static int DecrementRunningFilters(void);
static sfsistat mfclose(SMFICTX *ctx);
static int MXCheckFreeSlaves(void);
static int MXScanDir(char const *dir);
static int safeWriteHeader(FILE *fp, unsigned char const *str);

static char const *MultiplexorSocketName = NULL;
static char const *SubFilter = NULL;

static char const *FilterProgram = MIMEDEFANG_PL;

#define DATA ((struct privdata *) smfi_getpriv(ctx))

/* How many tries to get a unique directory name before giving up? */
#define MAXTRIES 8192

/* Size of buffer for temporary string manipulation */
#define SMALLBUF 2048

/* Size of chunk when replacing body */
#define CHUNK 4096

/* Number of file descriptors to close when forking */
#define CLOSEFDS 256

/* Number of running Perl filters */
static volatile int NumRunningFilters = 0;

/* Maximum allowable number of filters */
static int MaxRunningFilters = 10;

/* Mutex to protect access to NumRunningFilters */
static pthread_mutex_t FilterMutex = PTHREAD_MUTEX_INITIALIZER;

/* Do relay check? */
static int doRelayCheck = 0;

/* Do sender check? */
static int doSenderCheck = 0;

/* Keep directories around if multiplexor fails? */
static int keepFailedDirectories = 0;

#ifdef ENABLE_DEBUGGING
/**********************************************************************
*%FUNCTION: handle_sig
*%ARGUMENTS:
* s -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Handler for SIGSEGV and SIGBUS -- logs a message and returns -- hopefully,
* we'll get a nice core dump the second time around
***********************************************************************/
static void
handle_sig(int s)
{
    syslog(LOG_ERR, "WHOA, NELLY!  Caught signal %d -- this is bad news.  Core dump at 11.", s);

    /* Default is terminate and core. */
    signal(s, SIG_DFL);

    /* Return and probably cause core dump */
}
#endif

/**********************************************************************
* %FUNCTION: malloc_with_log
* %ARGUMENTS:
*  size -- amount of memory to allocate
* %RETURNS:
*  Allocated memory
* %DESCRIPTION:
*  Calls malloc, but syslogs an error on failure to allocate memory.
***********************************************************************/
static void *
malloc_with_log(size_t s)
{
    void *p = malloc(s);
    if (!p) {
	syslog(LOG_WARNING, "Failed to allocate %lu bytes of memory",
	       (unsigned long) s);
    }
    return p;
}

/**********************************************************************
* %FUNCTION: strdup_with_log
* %ARGUMENTS:
*  s -- string to strdup
* %RETURNS:
*  A copy of s in malloc'd memory.
* %DESCRIPTION:
*  Calls strdup, but syslogs an error on failure to allocate memory.
***********************************************************************/
static char *
strdup_with_log(char const *s)
{
    char *p = strdup(s);
    if (!p) {
	syslog(LOG_WARNING, "Failed to allocate %d bytes of memory in strdup",
	       (int) strlen(s)+1);
    }
    return p;
}

/**********************************************************************
* %FUNCTION: log_perl_stderr
* %ARGUMENTS:
*  ctx -- filter context
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  If the STDERR file exists, log its contents.
***********************************************************************/
static void
log_perl_stderr(SMFICTX *ctx)
{
    char buffer[SMALLBUF];
    struct privdata *data = DATA;
    FILE *fp;

    if (!data) return;

    /* If STDERR file is there, log contents */
    snprintf(buffer, SMALLBUF, "%s/STDERR", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	*buffer = 0;
	while(fgets(buffer, SMALLBUF, fp)) {
	    chomp(buffer);
	    if (*buffer) {
		syslog(LOG_DEBUG, "Perl stderr: %s", buffer);
	    }
	    *buffer = 0;
	}
	fclose(fp);
    }
}

/**********************************************************************
*%FUNCTION: closefiles
*%ARGUMENTS:
* None
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Hack -- closes a whole bunch of file descriptors.  Use after a fork()
*         to conserve descriptors.
***********************************************************************/
static void
closefiles(void)
{
    int i;
    for (i=0; i<CLOSEFDS; i++) {
	(void) close(i);
    }
}

/**********************************************************************
* %FUNCTION: convert_newline_sequences
* %ARGUMENTS:
*  str -- string to modify
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Converts the two-character sequence "\n" into literal newlines.
***********************************************************************/
static void
convert_newline_sequences(char *str)
{
    char *s = str;
    char *t = s;
    while(*s) {
	if (*s == '\\' && *(s+1) == 'n') {
	    s += 2;
	    *t++ = '\n';
	} else {
	    *t++ = *s++;
	}
    }
    *t = 0;
}

/**********************************************************************
*%FUNCTION: chomp
*%ARGUMENTS:
* str -- a string
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Removes newlines and carriage-returns (if any) from str
***********************************************************************/
static void
chomp(char *str)
{
    char *s, *t;
    s = str;
    for (t=str; *t; t++) {
	if (*t == '\n' || *t == '\r') continue;
	*s++ = *t;
    }
    *s = 0;
}

/**********************************************************************
* %FUNCTION: safeWriteHeader
* %ARGUMENTS:
*  fp -- file to write to
*  str -- a string value
* %RETURNS:
*  0 if header seems OK; 1 if suspicious character found.
* %DESCRIPTION:
*  Writes "str" to FP with the following changes:
*    CRLF -> written as-is
*    LFCR -> written as CRLF
*    CR   -> written as CRLF
*    LF   -> written as CRLF
***********************************************************************/
static int
safeWriteHeader(FILE *fp, unsigned char const *str)
{
    int suspicious = 0;
    for(; *str; str++) {
	/* ASCII-centric... */
	if (*str < ' ') {
	    /* CR is OK if followed by LF; tabs are OK; newlines by themselves
	       are OK (mail clients do that); anything else is suspicious */
	    switch(*str) {
	    case '\t':
	    case '\n':
		break;

	    case '\r':
		if (*(str+1) != '\n') {
		    suspicious = 1;
		}
		break;

	    default:
		/* ANY US-ASCII character is OK in header according
		   to RFC-2822 */
		break;
	    }
	}
#ifndef CONVERT_EMBEDDED_CRS_IN_HEADERS
	if (*str == '\r' && *(str+1) == '\n') {
	    fputc('\r', fp);
	    fputc('\n', fp);
	    str++;
	} else {
	    fputc(*str, fp);
	}
#else
	if (*str != '\r' && *str != '\n') {
	    fputc(*str, fp);
	    continue;
	}

	if (*str == '\r') {
	    if (*(str+1) == '\n') {
		str++;
	    }
	} else if (*str == '\n') {
	    if (*(str+1) == '\r') {
		str++;
	    }
	}
	fputc('\r', fp);
	fputc('\n', fp);
#endif
    }
    return suspicious;
}

/**********************************************************************
*%FUNCTION: IncrementRunningFilters
*%ARGUMENTS:
* None
*%RETURNS:
* -1 if we've hit maximum number of filters, 0 otherwise.
*%DESCRIPTION:
* Increments number of running filters using a mutex to lock access.
***********************************************************************/
static int
IncrementRunningFilters(void)
{
    if (pthread_mutex_lock(&FilterMutex)) {
	DEBUG(syslog(LOG_INFO, "IncrementRunningFilters: error locking FilterMutex"));
	return -1;
    }
    if (NumRunningFilters >= MaxRunningFilters) {
	syslog(LOG_INFO,
	       "Maximum number of filters (%d) running", MaxRunningFilters);
	(void) pthread_mutex_unlock(&FilterMutex);
	return -1;
    }
    NumRunningFilters++;
    DEBUG(syslog(LOG_INFO, "New filter; now running %d filters", NumRunningFilters));
    (void) pthread_mutex_unlock(&FilterMutex);
    return 0;
}

/**********************************************************************
*%FUNCTION: DecrementRunningFilters
*%ARGUMENTS:
* None
*%RETURNS:
* -1 if there was an error; 0 otherwise.
*%DESCRIPTION:
* Decrements number of running filters using a mutex to lock access.
***********************************************************************/
static int
DecrementRunningFilters(void)
{
    if (pthread_mutex_lock(&FilterMutex)) {
	DEBUG(syslog(LOG_INFO, "DecrementRunningFilters: error locking FilterMutex"));
	return -1;
    }
    if (NumRunningFilters == 0) {
	/* Shouldn't happen! */
	DEBUG(syslog(LOG_ERR, "DecrementRunningFilters: NumRunningFilters is zero!"));
	(void) pthread_mutex_unlock(&FilterMutex);
	return -1;
    }
    NumRunningFilters--;
    DEBUG(syslog(LOG_INFO, "Filter finished; now running %d filters", NumRunningFilters));
    (void) pthread_mutex_unlock(&FilterMutex);
    return 0;
}

/**********************************************************************
*%FUNCTION: mfconnect
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* hostname -- name of connecting host
* sa -- socket address of connecting host
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Allocates a private data structure for tracking this connection
***********************************************************************/
static sfsistat
mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
{
    struct privdata *data;
    struct sockaddr_in *insa = (struct sockaddr_in *) sa;
    char *tmp;

    DEBUG_ENTER("mfconnect", __LINE__);

    /* Delete any existing context data */
    mfclose(ctx);

    /* If too many running filters, reject connection at this phase */
    if (MultiplexorSocketName) {
	int n = MXCheckFreeSlaves();
	if (n <= 0) {
	    syslog(LOG_WARNING, "mfconnect: No free slaves");
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    } else {
	if (NumRunningFilters >= MaxRunningFilters) {
	    syslog(LOG_WARNING, "mfconnect: Maximum number of filters (%d) running",
		   MaxRunningFilters);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    }

    data = malloc_with_log(sizeof *data);
    if (!data) {
	DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->hostname = NULL;
    data->hostip   = NULL;
    data->dir      = NULL;
    data->fp       = NULL;
    data->headerFP = NULL;
    data->recipFP  = NULL;
    data->headerPresent  = 0;
    data->suspiciousBody = 0;
    data->lastWasCR      = 0;
    data->filterFailed   = 0;

    /* Save private data */
    smfi_setpriv(ctx, data);
    if (hostname) {
	data->hostname = strdup_with_log(hostname);
	if (!data->hostname) {
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    } else {
	data->hostname = NULL;
    }

    /* Padding -- should be big enough for IPv6 addresses, but currently
       only support IPv4 -- see inet_ntoa below.*/
    if (!sa) {
	data->hostip = NULL;
    } else {
	data->hostip = malloc_with_log(65);
	if (!data->hostip) {
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	tmp = inet_ntoa(insa->sin_addr);
	if (!tmp) {
	    syslog(LOG_WARNING, "inet_ntoa failed: %m");
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	strncpy(data->hostip, tmp, 64);
	data->hostip[64] = 0;
    }

    data->dir = NULL;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;
    data->suspiciousBody = 0;
    data->lastWasCR = 0;

    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;

}

/**********************************************************************
*%FUNCTION: envfrom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* from -- list of arguments to "MAIL FROM:" SMTP command.
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Allocates a private data structure for tracking this message, and
* opens a temporary file for saving message body.
***********************************************************************/
static sfsistat
envfrom(SMFICTX *ctx, char **from)
{
    struct privdata *data;
    int fd;
    int tries;
    int success;
    char buffer[SMALLBUF];
    char buf2[SMALLBUF];
    FILE *fp;
    time_t now = time(NULL);
    unsigned long ulnow = (unsigned long) now;

    DEBUG_ENTER("envfrom", __LINE__);
    /* Get the private context */
    data = DATA;
    if (!data) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Check relaying and sender */
    if (MultiplexorSocketName) {
	if (doRelayCheck) {
	    int n = MXRelayOK(ctx, data->hostip, data->hostname);
	    if (n == 0) {
		/* We reject connections from this relay */
		DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_REJECT");
		return SMFIS_REJECT;
	    }
	    if (n < 0) {
		/* Something went wrong */
		DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
		return SMFIS_TEMPFAIL;
	    }
	}
	if (doSenderCheck) {
	    int n = MXSenderOK(ctx, from[0]);
	    if (n == 0) {
		/* We reject connections from this sender */
		DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_REJECT");
		return SMFIS_REJECT;
	    }
	    if (n < 0) {
		/* Something went wrong */
		DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
		return SMFIS_TEMPFAIL;
	    }
	}
    }

    /* We need a nice clean directory all our own.  If two
     * threads get the same name, one of the mkdir() calls will fail. */

    success = 0;
    for (tries = 0; tries < MAXTRIES; tries++) {
	snprintf(buffer, SMALLBUF, "%s/mdefang-%lX-%d",
		 SPOOLDIR,
		 ulnow, tries);
	if (!mkdir(buffer, 0700)) {
	    success = 1;
	    break;
	}

	/* Apparently, mkdir on Solaris 8 can fail with EBADF */
	if ((errno != EEXIST) && (errno != EBADF)) {
	    break;
	}
    }

    /* Could not create temp. directory */
    if (!success) {
	syslog(LOG_WARNING, "Could not create mdefang directory in %s: %m",
	       SPOOLDIR);
	mfclose(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    if (data->dir) free(data->dir);
    if (data->fp) fclose(data->fp);
    if (data->headerFP) fclose(data->headerFP);
    if (data->recipFP) fclose(data->recipFP);

    data->dir = NULL;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;
    data->headerPresent = 0;

    data->dir = strdup_with_log(buffer);

    if (!data->dir) {
	/* Don't forget to clean up directory... */
	rmdir(buffer);
	mfclose(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Write host name and host IP */
    snprintf(buf2, SMALLBUF, "%s/HOSTNAME", buffer);
    fp = fopen(buf2, "w");
    if (fp) {
	if (data->hostname) {
	    fprintf(fp, "%s\n", data->hostname);
	} else {
	    fprintf(fp, "*UNKNOWN*\n");
	}
	fclose(fp);
    }
    snprintf(buf2, SMALLBUF, "%s/HOSTIP", buffer);
    fp = fopen(buf2, "w");
    if (fp) {
	if (data->hostip) {
	    fprintf(fp, "%s\n", data->hostip);
	} else {
	    fprintf(fp, "*UNKNOWN*\n");
	}
	fclose(fp);
    }

    /* Obtain and write the queue identifier */
    snprintf(buf2, SMALLBUF, "%s/QUEUEID", buffer);
    fp = fopen(buf2, "w");
    if (fp) {
	char *queueid = smfi_getsymval(ctx, "i");
	if (queueid) {
	    fprintf(fp, "%s\n", queueid);
	} else {
	    fprintf(fp, "\n");
	}
	fclose(fp);
    }

    /* Synthesize filename for output */
    snprintf(buf2, SMALLBUF, "%s/INPUTMSG", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->fp = fdopen(fd, "w+")) == NULL) {
	syslog(LOG_WARNING, "Could not create %s: %m", buffer);
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/HEADERS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->headerFP = fdopen(fd, "w+")) == NULL) {
	syslog(LOG_WARNING, "Could not create %s: %m", buffer);
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/RECIPIENTS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->recipFP = fdopen(fd, "w+")) == NULL) {
	syslog(LOG_WARNING, "Could not create %s: %m", buffer);
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Write the sender */
    snprintf(buf2, SMALLBUF, "%s/SENDER", data->dir);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(fp = fdopen(fd, "w+")) == NULL) {
	syslog(LOG_WARNING, "Could not create %s: %m", buffer);
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    fprintf(fp, "%s\n", from[0]);
    fclose(fp);

    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: rcptto
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* to -- list of arguments to each RCPT_TO
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Saves recipient data
***********************************************************************/
static sfsistat
rcptto(SMFICTX *ctx, char **to)
{
    DEBUG_ENTER("rcptto", __LINE__);
    if (!DATA) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("rcptto", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    fprintf(DATA->recipFP, "%s\n", to[0]);
    DEBUG_EXIT("rcptto", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: header
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* headerf -- Header field name
* headerv -- Header value
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes the header to the temporary file
***********************************************************************/
static sfsistat
header(SMFICTX *ctx, char *headerf, char *headerv)
{
    struct privdata *data = DATA;
    int suspicious = 0;

    DEBUG_ENTER("header", __LINE__);
    if (!data) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("header", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* We can close recipFP now to save a descriptor */
    if (data->recipFP && (fclose(data->recipFP) == EOF)) {
	data->recipFP = NULL;
	syslog(LOG_WARNING, "Error closing recipient descriptor: %m");
	DEBUG_EXIT("header", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->recipFP = NULL;

    /* Write the header to the message file */
    suspicious = safeWriteHeader(data->fp, (unsigned char const *) headerf);
    fprintf(data->fp, ": ");
    suspicious |= safeWriteHeader(data->fp, (unsigned char const *) headerv);
    fprintf(data->fp, "\r\n");
    if (suspicious) {
	char buffer[SMALLBUF];
	FILE *fp;
	snprintf(buffer, SMALLBUF, "%s/SUSPICIOUS-CHARS-IN-HEADERS",
		 data->dir);
	fp = fopen(buffer, "w");
	if (fp) fclose(fp);
    }

    /* Remove embedded newlines and save to our HEADERS file */
    chomp(headerf);
    chomp(headerv);
    fprintf(data->headerFP, "%s: %s\n", headerf, headerv);

    /* Check for existing identical X-Scanned-By: header */
    if (!strcasecmp(headerf, "X-Scanned-By") &&
	!strcasecmp(headerv, SCAN_BODY)) {
	data->headerPresent = 1;
    }

    DEBUG_EXIT("header", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eoh
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a blank line to indicate the end of headers.
***********************************************************************/
static sfsistat
eoh(SMFICTX *ctx)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("eoh", __LINE__);
    if (!data) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("eoh", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* We can close headerFP to save a descriptor */
    if (data->headerFP && (fclose(data->headerFP) == EOF)) {
	data->headerFP = NULL;
	syslog(LOG_WARNING, "Error closing header descriptor: %m");
	DEBUG_EXIT("eoh", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->headerFP = NULL;
    data->suspiciousBody = 0;
    data->lastWasCR = 0;

    fprintf(data->fp, "\r\n");
    DEBUG_EXIT("eoh", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: body
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* text -- a chunk of text from the mail body
* len -- length of chunk
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a chunk of the body to the temporary file
***********************************************************************/
static sfsistat
body(SMFICTX *ctx, u_char *text, size_t len)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("body", __LINE__);

    if (!data) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("body", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    if (fwrite(text, len, 1, data->fp) <= 0) {
	syslog(LOG_WARNING, "fwrite failed: %m line %d", __LINE__);
	cleanup(ctx);
	DEBUG_EXIT("body", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Scan body for suspicious characters */
    if (len) {
	u_char *s = text;
	size_t n;

	/* If last was CR, and this is not LF, suspicious! */
	if (data->lastWasCR && *text != '\n') {
	    data->suspiciousBody = 1;
	}

	data->lastWasCR = 0;
	for (n=0; n<len; n++, s++) {
	    if (*s == '\r') {
		if (n == len-1) {
		    data->lastWasCR = 1;
		} else if (*(s+1) != '\n') {
		    data->suspiciousBody = 1;
		}
		continue;
	    }

	    /* Control characters are OK in message body */
	    /*
	      if (*s < ' ' && *s != '\t' && *s != '\n') {
	         data->suspiciousBody = 1;
	      }
	    */
	    /* But embedded NULL's are cause for concern */
	    if (!*s) {
		data->suspiciousBody = 1;
	    }
	}
    }
    DEBUG_EXIT("body", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* This is where all the action happens.  Called at end of message, it
* forks the Perl scanner which may or may not ask for the body to be
* replaced.
***********************************************************************/
static sfsistat
eom(SMFICTX *ctx)
{
    pid_t n;
    int status;
    char buffer[SMALLBUF];
    FILE *fp;
    struct privdata *data = DATA;
    int r;
    int problem = 0;

    DEBUG_ENTER("eom", __LINE__);
    /* Close output file */
    if (!data) {
	syslog(LOG_WARNING, "Unable to obtain private data from milter context");
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    if (!data->fp || (fclose(data->fp) == EOF)) problem = 1;
    if (data->headerFP && (fclose(data->headerFP) == EOF)) problem = 1;
    if (data->recipFP && (fclose(data->recipFP) == EOF)) problem = 1;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    if (problem) {
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Create a sanity-check file */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if ((fp = fopen(buffer, "w")) != NULL) {
	fclose(fp);
    }

    /* Create SUSPICIOUS-CHARS-IN-BODY if needed */
    if (data->suspiciousBody) {
	snprintf(buffer, SMALLBUF, "%s/SUSPICIOUS-CHARS-IN-BODY", data->dir);
	if ((fp = fopen(buffer, "w")) != NULL) {
	    fclose(fp);
	}
    }

    data->suspiciousBody = 0;
    data->lastWasCR = 0;

    /* Run the Perl filter */
    if (MultiplexorSocketName) {
	if (MXScanDir(data->dir) < 0) {
	    data->filterFailed = 1;
	    /* If STDERR file is there, log contents */
	    log_perl_stderr(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    } else {
	if (IncrementRunningFilters() < 0) {
	    /* Doh!  Too many running filters... Sorry, have to abandon at
	       this point */
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	n = fork();
	if (n < 0) {
	    syslog(LOG_WARNING, "fork() failed -- cannot run filter");
	    DecrementRunningFilters();
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	} else if (n == 0) {
	    /* In the child */
	    char const *ppath = FilterProgram;
	    char const *pname = strrchr(ppath, '/');
	    if (pname) {
		pname++;
	    } else {
		pname = ppath;
	    }
	    closefiles();

	    /* Direct stdin/stdout to /dev/null; stderr to STDERR
	       in work dir */
	    open("/dev/null", O_RDWR);
	    open("/dev/null", O_RDWR);

	    snprintf(buffer, SMALLBUF, "%s/STDERR", data->dir);
	    open(buffer, O_WRONLY|O_CREAT|O_TRUNC);
	    if (SubFilter) {
		execl(ppath, pname, "-f", SubFilter, data->dir, NULL);
	    } else {
		execl(ppath, pname, data->dir, NULL);
	    }

	    /* No good to try to use syslog -- closefiles closed
	       the log descriptor.  Instead print to stderr and
	       parent will pick it up and log it */
	    fprintf(stderr, "Could not execute %s: %s\n",
		    ppath, strerror(errno));
	    _exit(EXIT_FAILURE);
	} else {
	    /* In the parent */
	    if (waitpid(n, &status, 0) < 0) {
		DecrementRunningFilters();
		cleanup(ctx);
		DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
		return SMFIS_TEMPFAIL;
	    }
	    if (DecrementRunningFilters() < 0) {
		DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
		return SMFIS_TEMPFAIL;
	    }
	}
    }

    /* If STDERR file is there, log contents */
    log_perl_stderr(ctx);

    /* If BOUNCE file is there, bounce the message */
    snprintf(buffer, SMALLBUF, "%s/BOUNCE", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	syslog(LOG_DEBUG, "Rejecting because filter instructed us to");
	buffer[0] = 0;
	fgets(buffer, SMALLBUF, fp);
	chomp(buffer);
	fclose(fp);
	smfi_setreply(ctx, "554", "5.7.1", buffer);
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_REJECT");
	return SMFIS_REJECT;
    }

    /* If DISCARD file is there, silently discard the message */
    snprintf(buffer, SMALLBUF, "%s/DISCARD", data->dir);
    if (!access(buffer, F_OK)) {
	syslog(LOG_DEBUG, "Discarding because filter instructed us to");
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_DISCARD");
	return SMFIS_DISCARD;
    }

    /* If TEMPFAIL file is there, tempfail the message */
    snprintf(buffer, SMALLBUF, "%s/TEMPFAIL", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	syslog(LOG_DEBUG, "Tempfailing because filter instructed us to");
	buffer[0] = 0;
	fgets(buffer, SMALLBUF, fp);
	chomp(buffer);
	fclose(fp);

	/* For compatibility with old versions of action_tempfail */
	if (!*buffer) {
	    strcpy(buffer, "Please try again later");
	}
	smfi_setreply(ctx, "451", "4.7.1", buffer);
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* If body is changed, replace it */
    snprintf(buffer, SMALLBUF, "%s/CHANGED", data->dir);
    if (!access(buffer, F_OK)) {
	int fd;
	int j;
	unsigned char chunk[CHUNK];
	snprintf(buffer, SMALLBUF, "%s/NEWBODY", data->dir);
	fd = open(buffer, O_RDONLY);
	if (fd < 0) {
	    syslog(LOG_WARNING, "Could not open %s for reading: %m", buffer);
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	while ((j=read(fd, chunk, CHUNK)) > 0) {
	    smfi_replacebody(ctx, chunk, j);
	}
	close(fd);
    }

    /* Do we need to change the Content-Type header? WARNING: WE CAN
       ONLY CHANGE A SINGLE-PART MESSAGE TO MULTIPART, BUT THE C CODE DOES
       NOT ENFORCE THIS! */
    snprintf(buffer, SMALLBUF, "%s/CONTENT-TYPE", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
#if SMFI_VERSION == 1
	/* Cannot change headers in Sendmail 8.10 -- please upgrade
	   to 8.11 */
#warning Please upgrade to Sendmail 8.11 -- Sendmail 8.10 with MIMEDefang will not handle single-part messages correctly.
#else
	fgets(buffer, SMALLBUF, fp);
	chomp(buffer);
	if (strlen(buffer) > 0) {
	    smfi_chgheader(ctx, "Content-Type", 1, buffer);
	}
	smfi_chgheader(ctx, "MIME-Version", 1, "1.0");
	/* We've just changed the message to multi-part.  Better delete
	   other MIME headers which may be present and confuse mail clients */
	smfi_chgheader(ctx, "Content-Disposition", 1, "");
	smfi_chgheader(ctx, "Content-Transfer-Encoding", 1, "");
#endif
	fclose(fp);
    }
    /* If there are new headers, add them */
    snprintf(buffer, SMALLBUF, "%s/NEWHEADERS", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	char header[SMALLBUF];
	char value[SMALLBUF];
	while (fgets(header, SMALLBUF, fp) && fgets(value, SMALLBUF, fp)) {
	    chomp(header);
	    chomp(value);
	    convert_newline_sequences(value);
	    smfi_addheader(ctx, header, value);
	}
	fclose(fp);
    }

    /* If there are headers to be changed, change them */
    snprintf(buffer, SMALLBUF, "%s/CHGHEADERS", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	char header[SMALLBUF];
	char value[SMALLBUF];
	while (fgets(header, SMALLBUF, fp) && fgets(value, SMALLBUF, fp)) {
	    chomp(header);
	    chomp(value);
	    convert_newline_sequences(value);
	    smfi_chgheader(ctx, header, 1, value);
	}
	fclose(fp);
    }

    /* If NEWRCPTS is there, add recipients */
    snprintf(buffer, SMALLBUF, "%s/NEWRCPTS", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	while (fgets(buffer, SMALLBUF, fp)) {
	    chomp(buffer);
	    smfi_addrcpt(ctx, buffer);
	}
	fclose(fp);
    }

    /* If DELRCPTS is there, delete recipients */
    snprintf(buffer, SMALLBUF, "%s/DELRCPTS", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	while (fgets(buffer, SMALLBUF, fp)) {
	    chomp(buffer);
	    smfi_delrcpt(ctx, buffer);
	}
	fclose(fp);
    }

    /* If UNFINISHED file is still there, ooops... */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if (!access(buffer, F_OK)) {
	syslog(LOG_WARNING, "Filter did not remove UNFINISHED file -- returning tempfail code to sendmail");
	data->filterFailed = 1;
	cleanup(ctx);
	DEBUG(syslog(LOG_INFO, "Exiting eom (line %d) ret=%d", __LINE__, SMFIS_TEMPFAIL));
	return SMFIS_TEMPFAIL;
    }

    if (!data->headerPresent) {
	smfi_addheader(ctx, "X-Scanned-By", SCAN_BODY);
    }
    r = cleanup(ctx);
    DEBUG(syslog(LOG_INFO, "Exiting eom (line %d) ret=%d", __LINE__, r));
    return r;
}

/**********************************************************************
*%FUNCTION: mfclose
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_ACCEPT
*%DESCRIPTION:
* Called when connection is closed.
***********************************************************************/
static sfsistat
mfclose(SMFICTX *ctx)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("mfclose", __LINE__);
    cleanup(ctx);
    if (data) {
	if (data->dir) free(data->dir);
	if (data->fp) fclose(data->fp);
	if (data->headerFP) fclose(data->headerFP);
	if (data->recipFP) fclose(data->recipFP);
	if (data->hostname) free(data->hostname);
	if (data->hostip) free(data->hostip);
	free(data);
    }
    smfi_setpriv(ctx, NULL);
    DEBUG_EXIT("mfclose", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: mfabort
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Called if current message is aborted.  Just cleans up.
***********************************************************************/
static sfsistat
mfabort(SMFICTX *ctx)
{
    return cleanup(ctx);
}

/**********************************************************************
*%FUNCTION: cleanup
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Cleans up temporary files.
***********************************************************************/
static sfsistat
cleanup(SMFICTX *ctx)
{
    sfsistat r = SMFIS_CONTINUE;
    struct privdata *data = DATA;
    pid_t n;
    int doRemove = 1;

    DEBUG_ENTER("cleanup", __LINE__);
    if (!data) {
	DEBUG_EXIT("cleanup", __LINE__, "SMFIS_CONTINUE");
	return r;
    }

    if (data->fp && (fclose(data->fp) == EOF)) {
	syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__);
	r = SMFIS_TEMPFAIL;
    }
    data->fp = NULL;

    if (data->headerFP && (fclose(data->headerFP) == EOF)) {
	syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__);
	r = SMFIS_TEMPFAIL;
    }
    data->headerFP = NULL;

    if (data->recipFP && (fclose(data->recipFP) == EOF)) {
	syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__);
	r = SMFIS_TEMPFAIL;
    }
    data->recipFP = NULL;

    /* Remove temporary files unless "-d" command-line switch given or
       SPOOLDIR/DO-NOT-DELETE-WORK-DIRS exists.  Also, if "-k" command-line
       switch was given and the filter failed, do not remove directory. */
    if (DebugMode ||
	(access(SPOOLDIR "/DO-NOT-DELETE-WORK-DIRS", F_OK) == 0)) {
	doRemove = 0;
    }

    if (keepFailedDirectories && data->filterFailed) {
	if (data->dir) {
	    syslog(LOG_WARNING, "Filter failed.  Message is in %s",
		   data->dir);
	}
	doRemove = 0;
    }

    if (doRemove) {
	if (data->dir) {
	    n = fork();
	    if (n < 0) {
		syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__);
		r = SMFIS_TEMPFAIL;
	    } else if (n == 0) {
		/* In the child */
		closefiles();
		execl(RM, "rm", "-r", "-f", data->dir, NULL);
		_exit(EXIT_FAILURE);
	    } else {
		/* In the parent */
		if (waitpid(n, NULL, 0) < 0) {
		    syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__);
		    r = SMFIS_TEMPFAIL;
		}
	    }
	}
    }

    if (data->dir) {
	free(data->dir);
	data->dir = NULL;
    }

    DEBUG(syslog(LOG_INFO, "Exiting cleanup (line %d) ret=%s", __LINE__, (r == SMFIS_TEMPFAIL ? "SMFIS_TEMPFAIL" : "SMFIS_CONTINUE")));
    return r;
}

static struct smfiDesc filterDescriptor =
{
    "MIMEDefang-" VERSION,      /* Filter name */
    SMFI_VERSION,		/* Version code */

#if SMFI_VERSION == 1
    /* We can: add a header and may alter body and add/delete recipients*/
    SMFIF_MODHDRS|SMFIF_MODBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT,
#else
    /* We can: Add headers, alter body, add/delete recipients, alter headers */
    SMFIF_ADDHDRS|SMFIF_CHGBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT|SMFIF_CHGHDRS,
#endif

    mfconnect,			/* connection */
    NULL,			/* HELO */
    envfrom,			/* MAIL FROM: */
    rcptto,			/* RCPT TO: */
    header,			/* Called for each header */
    eoh,			/* Called at end of headers */
    body,			/* Called for each body chunk */
    eom,			/* Called at end of message */
    mfabort,			/* Called on abort */
    mfclose			/* Called on connection close */
};

/**********************************************************************
* %FUNCTION: usage
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing (exits)
* %DESCRIPTION:
*  Prints usage information
***********************************************************************/
static void
usage(void)
{
    fprintf(stderr, "mimedefang version %s\n", VERSION);
    fprintf(stderr, "Usage: mimedefang [options]\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -h                -- Print usage info and exit\n");
    fprintf(stderr, "  -n maxFilters     -- Maximum number of concurrent filters\n");
    fprintf(stderr, "  -U user           -- Run as user instead of root\n");
    fprintf(stderr, "  -p /path          -- Path to UNIX-domain socket for sendmail communication\n");
    fprintf(stderr, "  -d                -- Enable debugging (do not remove spool files)\n");
    fprintf(stderr, "  -k                -- Do not remove spool files if filter fails\n");
    fprintf(stderr, "  -m /path          -- Use multiplexor; use /path as UNIX-domain socket\n");
    fprintf(stderr, "  -f /dir/filter    -- Specify full path of filter program\n");
    fprintf(stderr, "  -F filter_path    -- Specify path to filter rules file\n");
    fprintf(stderr, "  -r                -- Do relay check before processing body (requires -m)\n");
    fprintf(stderr, "  -s                -- Do sender check before processing body (requires -m)\n");
    fprintf(stderr, "  -P file           -- Write process-ID of daemon to specified file\n");
    exit(EXIT_FAILURE);
}

/**********************************************************************
* %FUNCTION: main
* %ARGUMENTS:
*  argc, argv -- the usual suspects
* %RETURNS:
*  Whatever smfi_main returns
* %DESCRIPTION:
*  Main program
***********************************************************************/
int
main(int argc, char **argv)
{
    int c;
    int maxConn;
    pid_t i;
    char *pidfile = NULL;
    struct passwd *pw = NULL;

    /* Paranoia time */
    umask(077);

    /* Paranoia time II */
    if (getuid() != geteuid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run suid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    if (getgid() != getegid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run sgid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    /* Process command line options */
    while ((c = getopt(argc, argv, "hn:p:dm:f:srkP:F:U:")) != -1) {
	switch (c) {
	case 'U':
	    /* User to run as */
	    if (user) {
		free(user);
	    }
	    user = strdup(optarg);
	    if (!user) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'F':
	    if (SubFilter) {
		free((void *) SubFilter);
	    }
	    SubFilter = strdup(optarg);
	    if (!SubFilter) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'P':
	    /* Write our pid to this file */
	    if (pidfile != NULL) free(pidfile);

	    pidfile = strdup(optarg);
	    if (!pidfile) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'k':
	    keepFailedDirectories = 1;
	    break;
	case 's':
	    doSenderCheck = 1;
	    break;
	case 'r':
	    doRelayCheck = 1;
	    break;
	case 'h':
	    usage();
	    break;
	case 'f':
	    /* Filter program */
	    if (optarg[0] != '/') {
		fprintf(stderr, "-f: You must supply an absolute path for filter program\n");
		exit(EXIT_FAILURE);
	    }
	    FilterProgram = strdup(optarg);
	    if (!FilterProgram) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'm':
	    /* Multiplexor */
	    MultiplexorSocketName = strdup(optarg);
	    if (!MultiplexorSocketName) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'd':
	    DebugMode = 1;
	    break;

	case 'n':
	    if (optarg == NULL || sscanf(optarg, "%d", &maxConn) != 1) {
		fprintf(stderr, "Usage: mimedefang -n max_filters\n");
		exit(EXIT_FAILURE);
	    }
	    if (maxConn <= 0) {
		fprintf(stderr, "Usage: mimedefang -n max_filters (max_filters > 0)\n");
		exit(EXIT_FAILURE);
	    }
	    MaxRunningFilters = maxConn;
	    break;
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		(void) fprintf(stderr, "Illegal conn: %s\n",
			       optarg);
		exit(EXIT_FAILURE);
	    }
	    /* Remove socket from file system */
	    (void) remove(optarg);
	    if (smfi_setconn(optarg) != MI_SUCCESS) {
		fprintf(stderr, "Could not open connection %s: %s",
			optarg, strerror(errno));
		exit(EXIT_FAILURE);
	    }
	    break;
	default:
	    usage();
	    break;
	}
    }

    /* Look up user */
    if (user) {
	pw = getpwnam(user);
	if (!pw) {
	    fprintf(stderr, "Unknown user '%s'", user);
	    exit(EXIT_FAILURE);
	}
	if (drop_privs(user, pw->pw_uid, pw->pw_gid) < 0) {
	    fprintf(stderr, "Could not drop privileges: %s", strerror(errno));
	    exit(EXIT_FAILURE);
	}
	free(user);
    }

    /* Warn */
    if (!getuid() || !geteuid()) {
	fprintf(stderr,
		"WARNING: You are running mimedefang as root.\n"
		"Future versions of MIMEDefang may refuse to run as root.\n");
    }


    if (smfi_register(filterDescriptor) == MI_FAILURE) {
	fprintf(stderr, "smfi_register failed\n");
	exit(EXIT_FAILURE);
    }

    /* Daemonize */
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	/* parent */
	exit(EXIT_SUCCESS);
    }
    setsid();
    signal(SIGHUP, SIG_IGN);
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	exit(EXIT_SUCCESS);
    }

    /* Write pid */
    if (pidfile) {
	FILE *fp = fopen(pidfile, "w");
	if (fp) {
	    fprintf(fp, "%d\n", (int) getpid());
	    fclose(fp);
	}
	free(pidfile);
    }

    chdir("/");
    (void) closelog();
    closefiles();

    /* Direct stdin/stdout/stderr to /dev/null */
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);

    openlog("mimedefang", LOG_PID, LOG_MAIL);

    /* Warn in syslog also*/
    if (!getuid() || !geteuid()) {
	syslog(LOG_ERR, "WARNING: You are running mimedefang as root.  Future versions of MIMEDefang may refuse to run as root.");
    }

#ifdef ENABLE_DEBUGGING
    signal(SIGSEGV, handle_sig);
    signal(SIGBUS, handle_sig);
#endif

    return smfi_main();
}

/**********************************************************************
* %FUNCTION: MXCommand
* %ARGUMENTS:
*  cmd -- command to send
*  buf -- buffer for reply
*  len -- length of buffer
* %RETURNS:
*  0 if all went well, -1 on error.
* %DESCRIPTION:
*  Sends a command to the multiplexor and reads the answer back.
***********************************************************************/
static int
MXCommand(char const *cmd,
	  char *buf,
	  int len)
{
    int fd;
    struct sockaddr_un addr;
    int nleft, nwritten, nread;
    int n;

    fd = socket(AF_LOCAL, SOCK_STREAM, 0);
    if (fd < 0) {
	syslog(LOG_ERR, "MXCommand: socket: %m");
	return -1;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, MultiplexorSocketName, sizeof(addr.sun_path) - 1);

    if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
	syslog(LOG_ERR, "MXCommand: connect: %m");
	close(fd);
	return -1;
    }

    nwritten = 0;
    nleft = strlen(cmd);
    n = 0;
    while(nleft) {
	n = write(fd, cmd+nwritten, nleft);
	if (n < 0) {
	    if (errno == EINTR) continue;
	    break;
	}
	nwritten += n;
	nleft -= n;
    }
    if (n < 0) {
	syslog(LOG_ERR, "MXCommand: write: %m");
	close(fd);
	return -1;
    }

    /* Now read the answer */
    nread = 0;
    nleft = len;
    while(nleft) {
	n = read(fd, buf+nread, nleft);
	if (n <= 0) {
	    if (n < 0 && errno == EINTR) continue;
	    break;
	}
	nread += n;
	nleft -= n;
    }
    if (n < 0) {
	syslog(LOG_ERR, "MXCommand: read: %m");
	close(fd);
	return -1;
    }
    buf[nread] = 0;
    close(fd);
    return 0;
}

/**********************************************************************
* %FUNCTION: MXCheckFreeSlaves
* %ARGUMENTS:
*  None
* %RETURNS:
*  >0 if there are free slaves, 0 if all slaves are busy, -1 if there
*  was an error.
* %DESCRIPTION:
*  Queries multiplexor for number of free slaves.
***********************************************************************/
static int
MXCheckFreeSlaves(void)
{
    char ans[SMALLBUF];
    int slaves;

    if (MXCommand("free\n", ans, SMALLBUF-1) < 0) return -1;

    if (sscanf(ans, "%d", &slaves) != 1) return -1;
    return slaves;
}

/**********************************************************************
* %FUNCTION: MXRelayOK
* %ARGUMENTS:
*  ip -- relay IP address
*  name -- relay name
* %RETURNS:
*  1 if it's OK to accept connections from this host; 0 if not, -1 if error.
*  If connection is rejected, error message *may* be set.
***********************************************************************/
static int
MXRelayOK(SMFICTX *ctx,
	  char const *ip,
	  char const *name)
{
    char cmd[SMALLBUF];
    char ans[SMALLBUF];
    int len;

    if (!ip || !*ip) {
	ip = "UNKNOWN";
    }
    if (!name || !*name) {
	name = ip;
    }

    snprintf(cmd, SMALLBUF, "relayok %s %s\n", ip, name);
    if (MXCommand(cmd, ans, SMALLBUF-1) < 0) return -1;
    if (!strcmp(ans, "ok 1\n")) return 1;
    if (!strcmp(ans, "ok 0\n")) return 0;

    chomp(ans);

    /* If rejection message is supplied, set failure code and return 0 */
    len = strlen(ans);
    if (len >= 6 && !strncmp(ans, "ok 0 ", 5)) {
	smfi_setreply(ctx, "554", "5.7.1", ans+5);
	return 0;
    }

    if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
    syslog(LOG_ERR, "Error from multiplexor: %s", ans);
    return -1;

}

/**********************************************************************
* %FUNCTION: MXSenderOK
* %ARGUMENTS:
*  ctx -- milter context
*  sender -- sender's e-mail address
* %RETURNS:
*  1 if it's OK to accept messages from this sender; 0 if not, -1 if error.
*  If message is rejected, error message *may* be set.
***********************************************************************/
static int
MXSenderOK(SMFICTX *ctx,
	   char const *sender)
{
    char cmd[SMALLBUF];
    char ans[SMALLBUF];
    int len;

    if (!sender || !*sender) {
	sender = "UNKNOWN";
    }

    snprintf(cmd, SMALLBUF, "senderok %s\n", sender);
    if (MXCommand(cmd, ans, SMALLBUF-1) < 0) return -1;
    if (!strcmp(ans, "ok 1\n")) return 1;
    if (!strcmp(ans, "ok 0\n")) return 0;

    chomp(ans);

    /* If rejection message is supplied, set failure code and return 0 */
    len = strlen(ans);
    if (len >= 6 && !strncmp(ans, "ok 0 ", 5)) {
	smfi_setreply(ctx, "554", "5.7.1", ans+5);
	return 0;
    }

    if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
    syslog(LOG_ERR, "Error from multiplexor: %s", ans);
    return -1;
}

/**********************************************************************
* %FUNCTION: MXScanDir
* %ARGUMENTS:
*  dir -- directory to scan
* %RETURNS:
*  0 if scanning succeeded; -1 if there was an error.
* %DESCRIPTION:
*  Asks multiplexor to initiate a scan.
***********************************************************************/
static int
MXScanDir(char const *dir)
{
    char cmd[SMALLBUF];
    char ans[SMALLBUF];
    int len;

    snprintf(cmd, SMALLBUF, "scan %s\n", dir);
    if (MXCommand(cmd, ans, SMALLBUF-1) < 0) return -1;

    if (!strcmp(ans, "ok\n")) return 0;

    len = strlen(ans);
    if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
    syslog(LOG_ERR, "Error from multiplexor: %s", ans);
    return -1;
}

#ifdef FOR_FUTURE_RELEASE
/**********************************************************************
* %FUNCTION: write_percent_encoded
* %ARGUMENTS:
*  str -- a string
*  fp -- file to write to
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Writes a string to a file, escaping the following chars:
*  < ascii-32; > ascii-126, %, \, ', ".  These are replaced with "%XY"
*  where X and Y are hex digits.
* "
***********************************************************************/
void
write_percent_encoded(unsigned char const *str,
		      FILE *fp)
{
    int c;
    while ((c = *str++) != 0) {
	if (c < 32 || c > 126 || c == '%' || c == '\\' || c == '\'' || c == '"') {
	    fprintf(fp, "%%%02X", (unsigned int) c);
	} else {
	    fputc(c, fp);
	}
    }
}

/**********************************************************************
* %FUNCTION: percent_decode
* %ARGUMENTS:
*  buf -- a buffer with percent-encoded data
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Decodes buf IN PLACE.
***********************************************************************/
static void
percent_decode(unsigned char *buf)
{
    unsigned char *in = buf;
    unsigned char *out = buf;
    unsigned int val;

    while(*in) {
	if (*in == '%' && isxdigit(*(in+1)) && isxdigit(*(in+2))) {
	    sscanf(in+1, "%2x", &val);
	    *out++ = (unsigned char) val;
	    in += 3;
	    continue;
	}
	*out++ = *in++;
    }
    /* Copy zero */
    *out = 0;
}
#endif
